extends KinematicBody2D

export var Speed = 18
export var Weight = 1
export var MaxJump = 2.25
export var JumpCutoff = 0.5
export var MovementStrength = 0.5
export var MovementStrengthInAir = 0.05
export var MovementStrengthOnStop = 0.5
export var EdgeJumpMilliseconds = 100
export var FallingJumpToleranceMilliseconds = 200

const CHAR_WIDTH = 24
const CHAR_HEIGHT = 44
const CELL_WIDTH = 16
const CELL_HEIGHT = 16
const MS_IN_SECS = 1000
const WALK_ANIM_TOLERANCE = 10
const MOVE_DELTA = 10.00

var velocity := Vector2()
var edgeJumpTimer := Timer.new()
var jumpTimer := Timer.new()
var canEdgeJump := false
var isJumping := false
var snap := false

onready var max_jump_velocity = CHAR_HEIGHT * MaxJump + (ProjectSettings.get_setting("physics/2d/default_gravity") + Weight) * 1.7
onready var max_momentum : float = Speed * CHAR_WIDTH
onready var gravity_acceleration = ProjectSettings.get_setting("physics/2d/default_gravity") * ProjectSettings.get_setting("physics/2d/default_gravity_vector") * Weight

func _ready():
	add_child(edgeJumpTimer)
	edgeJumpTimer.one_shot = true
	edgeJumpTimer.process_mode = Timer.TIMER_PROCESS_PHYSICS
	add_child(jumpTimer)
	jumpTimer.one_shot = true
	jumpTimer.process_mode = Timer.TIMER_PROCESS_PHYSICS

# warning-ignore:unused_argument
func _input(event):
	if event.is_action_pressed("ui_open_pause"):
		get_tree().quit()
	
	if event.is_action_pressed("player_move_jump"):
		jumpTimer.start(float(FallingJumpToleranceMilliseconds)/MS_IN_SECS)

var xPos = position.x
var endGame := false

func _physics_process(delta):
	_handle_input()
	
	if isJumping and Input.is_action_just_released("player_move_jump") and velocity.y < 0:
		velocity.y *= JumpCutoff
	
	if jumpTimer.time_left > 0 and edgeJumpTimer.time_left > 0:
		_jump()
	
	if not is_on_floor(): 
		velocity += gravity_acceleration * delta
		if velocity.y > gravity_acceleration.y:
			velocity.y = gravity_acceleration.y
	
	velocity = move_and_slide_with_snap(velocity,
		Vector2(0, 10) if snap else Vector2(),
		Vector2.UP, true, 4, deg2rad(46), false)
	
	if is_on_floor():
		edgeJumpTimer.start(float(EdgeJumpMilliseconds)/MS_IN_SECS)
		velocity.y = 0
		isJumping = false
		snap = true
	
func _handle_input():
	var direction = int(Input.is_action_pressed("player_move_right"))-int(Input.is_action_pressed("player_move_left"))
	var moveStrength = MovementStrengthInAir if not is_on_floor() else (MovementStrengthOnStop if abs(direction) < 0.01 else MovementStrength)
	velocity.x = lerp(velocity.x, max_momentum * direction, moveStrength)
	if abs(max_momentum * direction - velocity.x) < MOVE_DELTA: velocity.x = max_momentum * direction
		
func _jump():
	jumpTimer.stop()
	edgeJumpTimer.stop()
	isJumping = true
	snap = false
	velocity.y = -max_jump_velocity
	
var on_floor := false
var on_floor_body
var floor_velocity

func new_is_on_floor() -> bool: return on_floor

func floored() -> bool: return new_is_on_floor() or is_on_floor()
	
func move_and_snap(p_linear_velocity: Vector2, p_snap: Vector2, p_floor_direction: Vector2, p_stop_on_slope: bool, p_max_slides: int, p_floor_max_angle: float, p_infinite_inertia: bool):
	var was_on_floor := floored();

	var ret := move_and_slide(p_linear_velocity, p_floor_direction, p_stop_on_slope, p_max_slides, p_floor_max_angle, p_infinite_inertia);
	if !was_on_floor || p_snap == Vector2(): return ret;

	var col := move_and_collide(p_snap, p_infinite_inertia, false, true);
	var gt := get_global_transform();

	if col != null:
		var travel := col.travel
		var apply := true
		if p_floor_direction != Vector2():
			if acos(p_floor_direction.normalized().dot(col.normal)) < p_floor_max_angle:
				on_floor = true;
				on_floor_body = col.collider;
				floor_velocity = col.collider_velocity;
				if p_stop_on_slope:
					# move and collide may stray the object a bit because of pre un-stucking,
					# so only ensure that motion happens on floor direction in this case.
					travel = p_floor_direction * p_floor_direction.dot(travel.snapped(Vector2(0.5,0.5)));
					
			else: apply = false;

		if apply:
			print("transform: ", gt.origin)
			gt.origin += travel;
			set_global_transform(gt);
	return ret;